"""
mtg_mana_simulator allows you to quickly simulate Magic: the Gathering plays from a single
player's perspective to estimate certain measures related to available mana throughout the game.
"""

from mtg_mana_simulator.ai import AI
from mtg_mana_simulator.card import Card
from mtg_mana_simulator.context import Context
from mtg_mana_simulator.experiment import Experiment
from mtg_mana_simulator.metric import Metric
from mtg_mana_simulator.sequence import Sequence
from mtg_mana_simulator.version import __version__
